/*                          
** Module   : BlockDefinition.java
** Abstract : stores block-related information for runtime processing
**
** Copyright (c) 2005-2012, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20050809   @22097 Created initial version, supporting
**                           block information for undo and retry.
** 002 GES 20060202   @24224 Added external flag to allow processing that
**                           is dependent upon the external procedure to
**                           be properly hooked.  This enables the proper 
**                           cleanup of resources like temp-tables, 
**                           frames and streams.
** 003 GES 20060202   @24246 Added init and iterate flags to allow 
**                           the blockSetup() method of TransactionManager
**                           to compensate for Java continue statements
**                           which otherwise bypass iterate() calls.
** 004 GES 20060314   @25059 Added toString().
** 005 GES 20060510   @26057 Added blockTop flag for use during iterate()
**                           processing.
** 006 GES 20060626   @27568 Added suppressError flag.
** 007 GES 20060804   @28428 Removed iterate and blockTop flags.
** 008 CA  20080207   @37025 Added block type field.
** 009 ECF 20080218   @37117 Integrated generics.
** 010 GES 20080505   @38208 Changed undoables to be an array for 
**                           performance.
** 011 GES 20080613   @38883 Added hadPause and endkeyRetry flags, the
**                           retryCount counter and the backupUndoList.
** 012 GES 20080625   @38957 Changed to BlockType enum.
** 013 GES 20080627   @38991 Added isUndoLevel flag.
** 014 CA  20081127   @40707 Added iterated flag - indicates if a block
**                           was iterated at least once.
** 015 GES 20081120   @40739 Added properties bitmap and interactions
**                           counter.
** 016 GES 20081210   @40865 Added rollup property.
** 017 GES 20090422   @41904 Converted to standard string formatting.
** 018 ECF 20090816   @43652 Added OutputParameterAssigner field.
** 019 CA  20121220          Added the Block instance to the block definition.
*/

package com.goldencode.p2j.util;

import java.util.*;

/**
 * Container of block information for Progress compatible rollback (undo)
 * commit and retry.  The members of this class are intentionally made
 * <code>public</code> since this is just a simple container.
 */
public class BlockDefinition
{
   /** Sprintf format specification for string representation. */
   private static final String spec = "[label = %s; type = %s; full = %b; "  +
                                      "trans_level = %s; external = %b; "    +
                                      "top_level = %b; loop = %b; "          +
                                      "loop_protection = %b; had_pause = %b" +
                                      "; endkey_retry = %b; next_or_leave =" +
                                      " %s; is_retry = %b; needs_retry = %b" +
                                      "; ilp_count = %d; pending_break = %b" +
                                      "; properties = '%s']";
   
   /** Label to identify this block. */
   public String label = null;

   /** Caches whether or not this block started a transaction. */
   public boolean full = false;
   
   /** Level of transaction support for this block. */
   public int level = TransactionManager.NO_TRANSACTION;
   
   /** Bitmap specifying the block properties active for this block. */
   public int props = TransactionManager.PROP_NONE;
   
   /** Defines if this block is derived from an external procedure. */
   public boolean external = false;
   
   /** Defines if this block is a top-level block (a method) or not. */
   public boolean topLevel = false;
   
   /** Defines if this block is a loop or not. */
   public boolean loop = false;
   
   /** Has this block iterated at least once. */
   public boolean iterated = false;
   
   /** This block's type. */
   public BlockType type = BlockType.UNKNOWN;
   
   /** The block to be executed. */
   public Block block = null;

   /** Defines if this iteration requires infinite loop protection. */
   public boolean loopProtection = true;
   
   /** Defines if this iteration has encountered a pause. */
   public boolean hadPause = false;
   
   /** Marks the current retry as one caused by an <code>ENDKEY</code>. */
   public boolean endkeyRetry = false;
   
   /** Counts the number of sequential retries converted to NEXT or LEAVE. */
   public int ilpCount = -1;
   
   /** Counter for the number of interactive stmts in the current block. */
   public int interactions = 0;
   
   /** Aggregate interactions counter when a client block exits. */
   public boolean rollup = true;
   
   /** Has this block initialized at least once. */
   public boolean init = false;
   
   /**
    * Defines if this iteration has a pending break which was generated by 
    * infinite loop protection during a retry.
    */
   public boolean pendingBreak = false;
   
   /** Defines if this iteration requires a retry. */
   public boolean retry = false;
   
   /** Defines if this iteration is a retry (versus the first pass). */
   public boolean wasRetried = false;
   
   /**
    * Determines if infinite loop protection converts retry into next or 
    * leave.
    */
   public boolean next = false;
   
   /** Defines if this block was undone. */
   public boolean rolledBack = false;
   
   /** Defines if this block was already committed. */
   public boolean committed = false;
   
   /** Defines if this blocks is or was the target of UNDO. */
   public boolean isUndoLevel = false;
   
   /** Defines if this block's caller requires errors to be suppressed. */
   public boolean suppressError = false;
   
   /** Stores the list of undoable objects. */
   public UndoablePair[] undoables = null;
   
   /**
    * Stores the list of commitable objects to be notified at successful
    * completion of this block.
    */
   public List<Commitable> commitables = new ArrayList<Commitable>();
   
   /**
    * Stores the list of finalizable objects to be notified when the flow
    * of control exist from this block in any manner.
    */
   public List<Finalizable> finalizables = new ArrayList<Finalizable>();
   
   /** Stores the backup registry of undoable objects. */
   public PartitionedArray<Undoable> backupUndoList = null;
   
   /** Stores the output parameter assigner */
   public OutputParameterAssigner outputParmAssigner = null;
   
   /**
    * Construct an instance to represent a new block.
    *
    * @param    label
    *           This scope's label.
    * @param    level
    *           The type of transaction processing this scope can participate
    *           in.  This must be set to one of the following
    *           <code>TransactionManager</code> constants:
    *           <code>NO_TRANSACTION, SUB_TRANSACTION or TRANSACTION</code>.
    * @param    props
    *           A bitfield storing block properties which will be a
    *           combination of <code>TransactionManager</code> constants:
    *           <code>PROP_NONE, PROP_ERROR, PROP_ENDKEY, PROP_STOP</code>
    *           and <code>PROP_QUIT</code>.
    * @param    full
    *           If <code>true</code> this scope starts a full transaction
    *           (this is the scope at which database commits will occur).
    * @param    external
    *           <code>true</code> if this block is associated with the
    *           external procedure (in Progress terms). This controls the
    *           cleanup of certain resources such as temp-tables, frames
    *           and streams.
    * @param    topLevel
    *           <code>true</code> if this block is a method.
    * @param    loop
    *           <code>true</code> if this block is a loop.
    * @param    next
    *           <code>true</code> if this infinite loop protection should
    *           convert a retry into a next (continue), otherwise the retry
    *           will be converted into a leave (break).
    * @param    type
    *           This block's type.
    * @param    block
    *           The block to be executed.
    */
   public BlockDefinition(String    label,
                          int       level,
                          int       props,
                          boolean   full,
                          boolean   external,
                          boolean   topLevel,
                          boolean   loop,
                          boolean   next,
                          BlockType type,
                          Block     block)
   {
      this.label    = label;
      this.level    = level;
      this.props    = props;
      this.full     = full;
      this.external = external;
      this.topLevel = topLevel;
      this.loop     = loop;
      this.next     = next;
      this.type     = type;
      this.block    = block;
   }
   
   /**
    * Render this block's data as a human readable string.
    *
    * @return   The human readable form of this block's data.
    */
   public String toString()
   {
      Object[] parms = new Object[]
      {
         label,
         type.toString(),
         full,
         TransactionManager.decodeLevel(level),
         external,
         topLevel,
         loop,
         loopProtection,
         hadPause,
         endkeyRetry,
         (next ? "next" : "leave"),
         wasRetried,
         retry,
         ilpCount,
         pendingBreak,
         TransactionManager.decodeProperties(props)
      };
      
      return String.format(spec, parms);
   }
}
