/*
** Module   : HandleOps.java
** Abstract : Defines APIs to implement handle-specific operations.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 CA  20130112 Created initial version.
*/

package com.goldencode.p2j.util;

/**
 * Provides implementation for these handle-specific operations: 
 * <ul>
 * <li>DELETE OBJECT</li>
 * <li>DELETE PROCEDURE</li>
 * </ul>
 * 
 * Current implementation provides support only for SESSION, FILE-INFO, 
 * LAST-EVENT, ERROR-STATUS and procedure handles.
 */
public class HandleOps
{
   /**
    * Deletes the given handle. The current implementation allows only
    * procedure handles to be deleted.
    * <p>
    * In cases when the handle is a system handle, an {@link #invalidDelete()}
    * is called (to raise an error).
    * 
    * @param    h
    *           The handle to be deleted.
    */
   public static void delete(handle h)
   {
      if (h != null && h._isValid())
      {
         Object referent = h.get();
         if (h.isProcedure())
         {
            if (ProcedureManager.systemHandle(referent))
            {
               invalidDelete(h);
               return;
            }
   
            ProcedureManager.delete(h);
            return;
         }
         else
         {
            if (h.equals(SessionUtils.asHandle()) ||
                h.equals(FileSystemOps.asHandle()) ||
                h.equals(KeyReader.asHandle()) ||
                h.equals(ErrorManager.asHandle()))
            {
               invalidDelete();
               return;
            }
            
            throw new RuntimeException(
               "Delete implementation not available for handle " + 
               handle.unwrapType(h).getType().toStringMessage());
         }
      }
      
      invalidDelete();
   }

   /**
    * Invoked when the handle is invalid or a system (non-procedure) handle. 
    * Shows appropriate error messages and raises an error. 
    * Calls {@link #invalidDelete(handle) invalidDelete(null)}.
    */
   private static void invalidDelete()
   {
      invalidDelete(null);
   }

   /**
    * Invoked when the handle is invalid or a system (non-procedure) handle. 
    * Shows appropriate error messages and raises an error.
    * 
    * @param    h
    *           The handle for which the delete failed. May be 
    *           <code>null</code> if it was decide this is a non-procedure 
    *           handle.
    */
   private static void invalidDelete(handle h)
   {
      boolean procedure = h != null && h._isValid() && h.isProcedure();

      final String msg1 = "Invalid or inappropriate handle value given to " +
                          "DELETE OBJECT or DELETE PROCEDURE statement";
      final int id1 = 5425;

      final String msg2 = "Invalid or inappropriate handle value given to " +
                          "DELETE PROCEDURE statement. Procedure '%s':%d";
      final int id2 = 1428;
      // this can't be computed, is the legacy line num from the 4GL file.
      final int linenum = 0;
      
      int id = -1;
      String msg = null;
      
      if (procedure)
      {
         character pname = handle.unwrapProcedure(h).getFileName();
         msg = String.format(msg2, pname.toStringMessage(), linenum);
         id = id2;
      }
      else
      {
         msg = msg1;
         id = id1;
      }

      ErrorManager.recordOrThrowError(id, msg, false);
   }
}
